import * as React from 'react';
import PropTypes from 'prop-types';
import ScrollContainer from './ScrollContainer.js';
import MeasurementStore from './MeasurementStore.js';
import { DefaultLayoutSymbol, UniformRowLayoutSymbol } from './legacyLayoutSymbols.js';
import LegacyMasonryLayout from './layouts/MasonryLayout.js';
import LegacyUniformRowLayout from './layouts/UniformRowLayout.js';
declare type Layout = typeof DefaultLayoutSymbol | typeof UniformRowLayoutSymbol | LegacyMasonryLayout | LegacyUniformRowLayout;
interface Props<T> {
    columnWidth?: number;
    comp: React.ComponentType<{
        data: T;
        itemIdx: number;
        isMeasuring: boolean;
    }>;
    flexible?: boolean;
    gutterWidth?: number;
    items: T[];
    measurementStore?: any;
    minCols?: number;
    layout?: Layout;
    loadItems?: false | (({ from: number }: {
        from: number;
    }) => void | boolean | {});
    scrollContainer: () => HTMLElement | null;
    virtualBoundsTop?: number;
    virtualBoundsBottom?: number;
    virtualize?: boolean;
    className?: string;
    itemClassName?: string;
}
interface State<T> {
    hasPendingMeasurements: boolean;
    isFetching: boolean;
    items: T[];
    scrollTop: number;
    width?: number;
}
interface Position {
    top: number;
    left: number;
    width: number;
    height: number;
}
export default class Masonry<T> extends React.Component<Props<T>, State<T>> {
    static createMeasurementStore(): MeasurementStore<{}, {}>;
    /**
     * Delays resize handling in case the scroll container is still being resized.
     */
    handleResize: {
        (...args: any[]): void;
        clearTimeout(): void;
    };
    updateScrollPosition: {
        (...args: any[]): void;
        clearTimeout(): void;
    };
    measureContainerAsync: {
        (...args: any[]): void;
        clearTimeout(): void;
    };
    static propTypes: {
        /**
         * The preferred/target item width. If `flexible` is set, the item width will
         * grow to fill column space, and shrink to fit if below min columns.
         */
        columnWidth: PropTypes.Requireable<number>;
        /**
         * The component to render.
         */
        comp: PropTypes.Validator<(...args: any[]) => any>;
        /**
         * The preferred/target item width. Item width will grow to fill
         * column space, and shrink to fit if below min columns.
         */
        flexible: PropTypes.Requireable<boolean>;
        /**
         * The amount of space between each item.
         */
        gutterWidth: PropTypes.Requireable<number>;
        /**
         * An array of all objects to display in the grid.
         */
        items: PropTypes.Validator<(PropTypes.InferProps<{}> | null)[]>;
        /**
         * Measurement Store
         */
        measurementStore: PropTypes.Requireable<MeasurementStore<object, {}>>;
        /**
         * Layout system to use for items
         */
        layout: PropTypes.Requireable<LegacyMasonryLayout>;
        /**
         * A callback which the grid calls when we need to load more items as the user scrolls.
         * The callback should update the state of the items, and pass those in as props
         * to this component.
         */
        loadItems: PropTypes.Requireable<(...args: any[]) => any>;
        /**
         * Minimum number of columns to display.
         */
        minCols: PropTypes.Requireable<number>;
        /**
         * Function that the grid calls to get the scroll container.
         * This is required if the grid is expected to be scrollable.
         */
        scrollContainer: PropTypes.Requireable<(...args: any[]) => any>;
        /**
         * Whether or not to use actual virtualization
         */
        virtualize: PropTypes.Requireable<boolean>;
        /**
         * className
         */
        className: PropTypes.Requireable<string>;
        /**
         * itemClassName
         */
        itemClassName: PropTypes.Requireable<string>;
    };
    static defaultProps: {
        columnWidth: number;
        measurementStore: MeasurementStore<{}, {}>;
        minCols: number;
        layout: symbol;
        loadItems: () => void;
        virtualize: boolean;
        className: string;
        itemClassName: string;
    };
    constructor(props: Props<T>);
    /**
     * Adds hooks after the component mounts.
     */
    componentDidMount(): void;
    componentDidUpdate(_prevProps: Props<T>, prevState: State<T>): void;
    /**
     * Remove listeners when unmounting.
     */
    componentWillUnmount(): void;
    static getDerivedStateFromProps<U>(props: Props<U>, state: State<U>): {
        hasPendingMeasurements: boolean;
        items: U[];
        isFetching: boolean;
    } | {
        hasPendingMeasurements: boolean;
        items: U[];
        isFetching?: undefined;
    } | null;
    setGridWrapperRef: (ref: HTMLElement | null) => void;
    setScrollContainerRef: (ref: ScrollContainer | null) => void;
    fetchMore: () => void;
    containerHeight: number;
    containerOffset: number;
    gridWrapper: HTMLElement | null;
    insertAnimationFrame?: number;
    scrollContainer: ScrollContainer | null;
    measureContainer(): void;
    /**
     * Clear measurements/positions and force a reflow of the entire grid.
     * Only use this if absolutely necessary - ex: We need to reflow items if the
     * number of columns we would display should change after a resize.
     */
    reflow(): void;
    renderMasonryComponent: (itemData: T, idx: number, position: Position) => JSX.Element | null;
    render(): JSX.Element;
}
export {};
